#!@BIN_SH@
# @(#)$GW: cronrun/cronrun.sh,v 1.14 2015/05/26 19:36:27 kim Exp $
#
# 20001129  Kimmo Suominen <kimmo@suominen.com>
#
# Copyright (c) 2000-2015 Kimmo Suominen
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# We will use the caller's PATH for running the provided command line.
OPATH="${PATH}"

# This is the trusted path for this script.
PATH=/bin:/usr/bin:/usr/sbin:/usr/pkg/bin:/usr/pkg/sbin:/usr/lib
export PATH

PROG="${0##*/}"
CONFFILE="/etc/${PROG}.conf"

# Nagios exit codes.
EXIT_OK=0
EXIT_WARNING=1
EXIT_CRITICAL=2
EXIT_UNKNOWN=3

# Nagios server.
NAGIOS_SERVERS=
NSCA_PORT=5667
SEND_NSCA_CONFIG=/etc/send_nsca.cfg

###############################################################################

#
# Mail the temporary file if it is not empty.
#
mailit ()
{
    case "${charset}" in
    '')
	charset='us-ascii'
	encode=false
	;;
    us-ascii)
	encode=false
	;;
    *)
	encode=true
	;;
    esac

    # XXX: Fix header encoding.
    #
    # More than just subject should be encoded.  Also, the character set
    # in the headers is more limited than in the body, so the call to
    # recode is insufficient for accurate conformance to RFC 2047.
    #
    # In addition to the question mark and space, also the equal sign (=)
    # should be encoded. The "fgrep" below is not sufficient, it should be
    # better to just compare the original and encoded strings.
    #
    # http://en.wikipedia.org/wiki/MIME#Encoded-Word

    if ${encode}
    then
	es=$(echo "${subject}" | recode data..qp)
	if echo "${es}" | fgrep -q =
	then
	    es=$(echo "${es}" | sed -e 's/ /_/g' -e 's/\?/=3F/g')
	    subject="=?${charset}?Q?${es}?="
	fi
    fi

    (
	if [ "${from}" ]
	then
	    echo "From: ${from}"
	fi
	recipientsdone=false
	if [ "${errorsto}" ]
	then
	    if ${errors} || [ ${STATUS} -ne 0 ]
	    then
		echo "To: ${errorsto#, }"
		echo "Subject: FAILED (${STATUS}) - ${subject}"
		recipientsdone=true
	    fi
	fi
	if ! ${recipientsdone}
	then
	    echo "To: ${to#, }"
	    if [ "${cc}" ]
	    then
		echo "cc: ${cc#, }"
	    fi
	    if [ "${bcc}" ]
	    then
		echo "BCC: ${bcc#, }"
	    fi
	    echo "Subject: ${subject}"
	fi
	if [ ${STATUS} -ne 0 ]
	then
	    echo 'Importance: High'
	fi
	echo "Mime-Version: 1.0"
	echo "Content-Type: text/plain; charset=${charset}"
	if ${encode}
	then
	    echo "Content-Transfer-Encoding: quoted-printable"
	fi
	echo ''
	(
	    if ${errors}
	    then
		echo '=== ERROR OUTPUT ==='
		echo ''
		cat "${TMPERR}"
		echo ''
		echo '=== REPORT OUTPUT ==='
		echo ''
	    fi
	    cat "${TMPOUT}"
	) \
	| if ${encode}
	then
	    recode data..qp
	else
	    cat
	fi
	echo ''
    ) \
    | if ${verbose}
    then
	cat
    else
	sendmail -t
    fi
}

#
# Nagios notification.
#
nagiosit()
{
    code=${STATUS}
    case ${code} in
    ${EXIT_CRITICAL})
	tag=CRITICAL
	;;
    ${EXIT_WARNING})
	tag=WARNING
	;;
    ${EXIT_OK})
	tag=OK
	;;
    *)
	code=${EXIT_UNKNOWN}
	tag=UNKNOWN
	;;
    esac
    if ${errors}
    then
	statusinfo="${tag}: $(head -1 "${TMPERR}")"
    elif ${output}
    then
	statusinfo="${tag}: $(head -1 "${TMPOUT}")"
    else
	statusinfo="${tag}"
    fi
    nsca_data=$(printf "${hostname}\t${service}\t${code}\t${statusinfo}\n")
    if ${verbose}
    then
	echo "${nsca_data}" | tr '\t' :
	nagios_status=${EXIT_OK}
    else
	(
	    if ${quiet}
	    then
		exec 3>/dev/null
	    else
		exec 3>&1
	    fi
	    for nagios_server in $(echo "${NAGIOS_SERVERS}" | tr ' ,' '\n\n')
	    do
		echo "${nsca_data}" \
		| send_nsca \
		    -H "${nagios_server}" \
		    ${NSCA_PORT:+-p ${NSCA_PORT}} \
		    ${SEND_NSCA_CONFIG:+-c "${SEND_NSCA_CONFIG}"} \
		    1>&3 2>&1
	    done
	)
    fi
}

#
# Run the command using the caller's PATH.
#
runit()
{
    (export PATH; PATH="${OPATH}"; "${@}")
    STATUS=${?}
}

###############################################################################

#
# Die with an error.
#
die()
{
    echo "${PROG}: ${@}" 1>&2
    exit ${EXIT_CRITICAL}
}

#
# Output a warning.
#
warn()
{
    echo "${PROG}: ${@}" 1>&2
}

###############################################################################

#
# Output usage information.
#
usage()
{
    cat <<-EOF
	Usage:  ${PROG} <options> command [args ...]

	Run a command and report its status via mail or by posting to Nagios.
	You should specify either the -m or the -n option or both.

	Mail mode:
	  -m		Mail command output
	  -t to		Recipients
	  -e errorsto	Recipients for a failed run (one with errors)
	  -c cc		Carbon copy (visible recipients)
	  -b bcc	Blind carbon copy (invisible recipients)
	  -f from	Sender
	  -s subject	Subject
	  -T		Prepend hostname to subject (see also -H)
	  -C charset	Encode the message, label with this character set

	Report mode:
	  -r		Report mode: separate errors from output (see manual)

	  All mail mode options are applicable to report mode.

	Nagios mode:
	  -n		Send status to Nagios (NSCA)
	  -H host	Host to use in status message
	  -S service	Service to use in status message
	  -N host	Nagios server host (receiver of NCSA messages)
	  -P port	NSCA service port number

	General options:
	  -d		Dump configuration, do not run anything (debug)
	  -p logdir	Path to the log directory
	  -q		Silence some errors (quiet; see manual)
	  -v		Output all status information on stdout only (verbose)
	  -w maxwait	Wait up to maxwait seconds before running command

	The command is required (of course) and any remaining arguments are
	passed along to it.

	Default values can be set and changed in the configuration file.
	EOF
}

dumpconf()
{
    cat <<-EOF
	General settings:
	- configuration file: ${CONFFILE} (${CONFSTATUS})
	- path to log directory: ${logdir}
	- logfile basename: ${logbase}
	EOF
    if ${mail}
    then
	cat <<-EOF

	Mail settings:
	- From: ${from}
	- Errors To: ${errorsto#, }
	- To: ${to#, }
	- cc: ${cc#, }
	- bcc: ${bcc#, }
	- Subject: ${subject}
	EOF
    fi
    if ${nagios}
    then
	cat <<-EOF

	Nagios settings:
	- send NSCA notifications to port ${NSCA_PORT} on
	$(echo "${NAGIOS_SERVERS}" | tr ' ,' '\n\n' | sed -e 's/^/  - /')
	- configuration for send_nsca: ${SEND_NSCA_CONFIG}
	- post status for: ${hostname}, service ${service}
	EOF
    fi
}

###############################################################################

if [ -z "${USER}" ]
then
    USER="${LOGNAME:-$(id -un)}"
fi
if [ -z "$(echo "${USER}" | tr -d 0-9)" ]
then
    die 'Cannot determine username of effective user'
fi

debug=false
hostname=$(netname)
legacy=true
mail=false
maxwait=0
nagios=false
report=false
tagsubject=false
verbose=false

if [ -t 0 ]
then
    quiet=false
else
    # Avoid errors from send_nsca being mailed back by cron.
    quiet=true
fi

if [ -f "${CONFFILE}" ]
then
    if [ -r "${CONFFILE}" ]
    then
	. "${CONFFILE}"
	CONFSTATUS="found, readable"
    else
	CONFSTATUS="found, unreadable"
    fi
else
    CONFSTATUS="not found"
fi

#
# Support for legacy configuration variables.
#
NAGIOS_SERVERS="${NAGIOS_SERVERS:-${NAGIOS_SERVER}}"

while getopts C:H:N:P:S:Tb:c:de:f:hmnp:qrs:t:vw: opt
do
    legacy=false
    case "${opt}" in
    C) charset=$(echo "${OPTARG}" | tr 'A-Z' 'a-z');;
    H) hostname="${OPTARG}";;
    N) NAGIOS_SERVERS="${OPTARG}";;
    P) NSCA_PORT=$(echo "${OPTARG}" | tr -cd '0-9');;
    S) service="${OPTARG}";;
    T) tagsubject=true;;
    b) bcc="${bcc}, ${OPTARG}";;
    c) cc="${cc}, ${OPTARG}";;
    d) debug=true;;
    e) errorsto="${errorsto}, ${OPTARG}";;
    f) from="${OPTARG}";;
    h) usage; exit ${EXIT_OK};;
    m) mail=true;;
    n) nagios=true;;
    q) quiet=true;;
    p) logdir="${OPTARG}";;
    r) report=true;;
    s) subject="${OPTARG}";;
    t) to="${to}, ${OPTARG}";;
    v) verbose=true;;
    w) maxwait=$((${OPTARG} + 0));;
    *) usage 1>&2; exit ${EXIT_CRITICAL};;
    esac
done
shift $((${OPTIND} - 1))

if ${legacy}
then
    if [ ${#} -lt 3 ]
    then
	usage 1>&2
	exit ${EXIT_CRITICAL}
    fi
    subject="${1}"
    tagsubject=true
    to="${2}"
    shift 2
fi

if ${report}
then
    mail=true
    if [ -z "${errorsto}" ]
    then
	errorsto="${USER}"
    fi
fi

if [ -z "${bcc}" -a -z "${cc}" -a -z "${to}" ]
then
    bcc="${USER}"
fi

if [ -z "${to}" ]
then
    to='Undisclosed:;'
fi

if ${nagios}
then
    if [ -z "${NAGIOS_SERVERS}" ]
    then
	die 'Must specify at least one nagios server'
    fi
fi

if [ -z "${service}" ]
then
    for i
    do
	case "${i}" in
	*=*)
	    # Skip variable assignments.
	    ;;
	*)
	    service="${i##*/}"
	    break
	    ;;
	esac
    done
fi
if [ -z "${service}" ]
then
    die 'Cannot extract service name from command'
fi

if [ -z "${subject}" ]
then
    subject="${hostname}: ${service}"
elif ${tagsubject}
then
    subject="${hostname}: ${subject}"
fi

if ! ${mail} && ! ${nagios}
then
    mail=true
fi

if [ -z "${logdir}" ]
then
    if ${mail}
    then
	logdir="${TMPDIR:-/tmp}"
    else
	logdir="/var/log/${PROG}"
    fi
fi

if [ ${#} -lt 1 ]
then
    die 'Must specify command to run'
fi

TIMESTAMP=$(date '+%Y-%m-%d.%H:%M')
if [ -z "${TIMESTAMP}" ]
then
    die 'Cannot create timestamp'
fi
logbase="${service}.${USER}.${TIMESTAMP}"

if ${debug}
then
    dumpconf
    cat <<-EOF

	Command:
	${@}
	EOF
    exit ${EXIT_OK}
fi

#
# Setup output files.
#
# The /var/log/${PROG} directory should writable by all and have its
# sticky bit set, and created at system boot if missing.
#
TMPOUT=$(mktemp "${logdir}/${logbase}.o.XXXXXXXX") || exit ${EXIT_CRITICAL}
TMPERR=$(mktemp "${logdir}/${logbase}.e.XXXXXXXX") || exit ${EXIT_CRITICAL}
if ! ${nagios}
then
    trap "rm -f '${TMPOUT}' '${TMPERR}'" 0 1 3 15
fi

#
# Sleep.
#
if [ ${maxwait} -gt 0 ]
then
    waitsec=$(awk -vmin=0 -vmax=${maxwait} \
	'BEGIN { srand(); print int(min + rand() * (max - min + 1)) }'
    )
    if ${verbose}
    then
	warn "Would have slept for ${waitsec} of ${maxwait} seconds."
    else
	sleep "${waitsec}"
    fi
fi

#
# Run the command using the caller's PATH.
#
if ${report}
then
    runit "${@}" 1>"${TMPOUT}" 2>"${TMPERR}"
else
    runit "${@}" 1>"${TMPOUT}" 2>&1
fi

errors=false
if [ -s "${TMPERR}" ]
then
    errors=true
fi

results=false
if [ -s "${TMPOUT}" ]
then
    results=true
fi

#
# Mail the output.
#
if ${mail}
then
    if ${errors} || ${results}
    then
	mailit
    else
	if ${verbose}
	then
	    warn 'No output from command!'
	fi
    fi
fi

#
# Post to nagios.
#
if ${nagios}
then
    nagiosit
fi

#
# Preserve exit status.
#
exit ${STATUS}
